from enum import Enum
from pathlib import Path
from typing import Optional, Union, Dict

from pydantic import BaseModel, FilePath, AnyUrl

from sito_io.errors import UnrootedResource


class UriKind(Enum):
    """Describes the kind of URI. URIs exist on a continuum, from simply a uniform identifier,
    to a fully-qualified URL.
    """
    Naive = 0  # we don't know anything really, it's just some identifier
    Urn = 1  # it's a Uniform Resource Name, the red-headed stepchild of the URI family
    Relative = 2  # it's a relative URI, we can resolve iff we have context
    Absolute = 3  # it's an absolute path, with the implication of locality, but not necessarily local file system
    FullyQualified = 4  # base is a full URI with scheme


class LocationKind(Enum):
    """Describes where to actually find the resource"""
    Naive = 0  # we don't know how it's stored
    Abstract = 1  # it's not really a tangible thing, more of a concept, such as a Graph
    LocalArtifact = 2  # it's a local file we can immediately open
    ArchiveArtifact = 3  # represents part of an archive tar/zip/etc
    Network = 4  # it's a local absolute path


class ResourceKind(Enum):
    """Describes what the resource actually is.
    Artifact is a file/directory, or blob of bytes. It is basically anything you could "rsync".
    """
    Naive = 0  # we don't know what it is yet
    Artifact = 1  # a file or directory, existing on some filesystem, somewhere.
    File = 2  # a file on a filesystem
    Directory = 3  # a directory on a filesystem


class Resource(BaseModel):
    """ A Resource is anything identified by a Uniform Resource Identifier.
    For those who don't like recursive definitions, a Resource is a logical or physical resource used by
    web technologies. URIs may be used to identify anything, including real-world objects, such as people and places,
    concepts, or information resources such as web pages and books.
    URIs are UNIFORM, not UNIVERSAL, meaning they have a specific structure. They may be fully qualified or relative.
    A URL is a uniform resource locator. All URLs are URIs, but not all URIs are URLs. Relative URIs and
    URNs are URIs without a strict locator, but they can be resolved to URLs.
    """

    id: Optional[str]  # some abstract identifier
    uri: Union[FilePath, AnyUrl, Path]
    type: str = ""  # mimetype
    kind: UriKind = UriKind.Naive
    attrs: Dict[str, str] = {}
    meta: str = ""

    def is_localized(self):
        """Path is resolvable as a local resource, NOT necessarily existing locally"""
        if self.kind in (UriKind.Absolute, UriKind.Relative):
            # todo: I dunno if I wanna deal with relative squirrelyness here
            return True

        elif self.kind is UriKind.FullyQualified:
            if self.uri.startswith('file://'):
                return True

    def __fspath__(self):
        if self.is_localized():
            return str(self.uri)
        raise UnrootedResource(f"Cannot resolve absolute path: {self.uri}")


from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union
from pydantic import BaseModel, FilePath, DirectoryPath, AnyUrl

from sito_io.dctypes.resource import UriKind, Resource

UriT = Union[FilePath, AnyUrl]
OptionsT = Optional[Union[Dict[str, str], List[Tuple[str, str]], List[str]]]
# todo: pending optional directories for output dirs that will be created,
# see: https://github.com/samuelcolvin/pydantic/issues/1254
# https://github.com/samuelcolvin/pydantic/issues/1983


class Manifest(BaseModel):
    """A loose collection of elements"""

    base: Union[DirectoryPath, AnyUrl, Path]  # base of the manifest tree
    kind: UriKind = UriKind.Naive  # how is the manifest rooted?
    elements: List[Resource]  # manifest contents


class ManifestMap(BaseModel):
    """A Manifest which can be addressed by the relative paths"""

    base: Union[DirectoryPath, AnyUrl, Path]  # base of the manifest tree
    kind: UriKind = UriKind.Naive  # how is the manifest rooted?
    elements: Dict[Path, Resource]  # manifest contents


class SitoFileToFile(BaseModel):
    """File-in, file-out. Takes a single file (or URI), emits a single file (or URI) """

    input_uri: UriT
    output_uri: Optional[Union[FilePath, AnyUrl, str]]
    options: OptionsT


class SitoDirToDir(BaseModel):
    """Directory-in, directory-out. """

    input_dir: DirectoryPath
    output_dir: Optional[Union[DirectoryPath, str]]
    options: OptionsT


class SitoFileToDir(BaseModel):
    """Single file in, directory of files out. """

    input_uri: UriT
    output_dir: Optional[Union[DirectoryPath, str]]
    options: OptionsT


class SitoCoreUtil(BaseModel):
    """Emulates the interface of common coreutils tools, mv, tar, etc. E.g.:
    tool [options] [output] <input> [inputs...]"""

    inputs: List[UriT]
    output: Optional[UriT]
    options: OptionsT



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A pure python implementation of the Resource interface, based on dataclasses
"""

import os
import json
from typing import Optional, Union, Dict
from enum import Enum
from urllib.parse import urlsplit, SplitResult
from pathlib import Path
import dataclasses


class ResourceKind(Enum):
    Naive = 0  # we don't know how it's rooted
    Relative = 1  # it's a relative tree, base may lack context
    Absolute = 2  # it's a local absolute path
    FullyQualified = 3  # base is a full URI with scheme
    # todo: relative http urls?


class LocationKind(Enum):
    Naive = 0  # we don't know how it's stored
    Local = 1  # it's a local file we can immediately open
    Archive = 2  # represents part of an archive tar/zip/etc
    Network = 3  # it's a network resource with a protocol
    NoProtoFile = 4  # it's a file but it is not local (no access scheme)


class UnrootedResource(Exception):
    ...


def kind_from_uriparts(parts: SplitResult) -> ResourceKind:
    if parts.scheme:
        return ResourceKind.FullyQualified
    elif os.path.isabs(parts.path):
        return ResourceKind.Absolute
    else:
        return ResourceKind.Relative


def location_kind_from_uriparts(parts: SplitResult) -> LocationKind:
    # todo: archive?
    if parts.scheme == "":
        return LocationKind.Local
    elif parts.scheme == "file":
        if parts.netloc in ("", "localhost"):
            return LocationKind.Local
        return LocationKind.NoProtoFile
    else:
        return LocationKind.Network


def path_from_file_uri(uri: str) -> str:
    """Convert a file URI to a local path. Idempotent on regular paths.
    Raises if presented with a scheme besides file://
    """
    if uri.startswith("file://"):
        return uri.replace("file://", "")

    if "://" in uri:
        raise ValueError(f"Cannot convert URI to path: {uri}")
    return uri


@dataclasses.dataclass(frozen=True)
class Resource(object):
    """Describes some local or remote resource"""

    uri: str
    id: Optional[str] = ""  # some abstract identifier
    type: str = ""  # mimetype
    attrs: Dict[str, str] = dataclasses.field(default_factory=dict)
    meta: str = ""
    kind: ResourceKind = dataclasses.field(default=ResourceKind.Naive, init=True)
    location_kind: LocationKind = dataclasses.field(
        default=LocationKind.Naive, init=True
    )
    _parts: SplitResult = dataclasses.field(init=False, repr=False)

    def __post_init__(self):
        if isinstance(self.uri, Resource):
            object.__setattr__(self, "uri", str(self.uri.uri))
        parts = urlsplit(self.uri)
        object.__setattr__(self, "_parts", parts)
        object.__setattr__(self, "kind", kind_from_uriparts(parts))
        object.__setattr__(self, "location_kind", location_kind_from_uriparts(parts))

    def __str__(self):
        return self.uri

    def retrieve(self, filename=None, reporthook=None, data=None) -> "LocalResource":
        import urllib

        path, meta = urllib.request.urlretrieve(
            self.uri, filename=filename, reporthook=reporthook, data=data
        )
        assert os.path.isfile(path)
        attrs = self.attrs.copy()
        attrs.update(
            {
                "user.origin.uri": self.uri,
                "user.charsets": json.dumps(meta.get_charsets()),
            }
        )
        path = "file://" + path
        return LocalResource(
            path,
            id=self.id,
            type=meta.get_content_type(),
            attrs=attrs,
            kind=ResourceKind.FullyQualified,
            location_kind=LocationKind.Local,
        )

    def is_localized(self):
        """Path is resolvable as a local resource, NOT necessarily existing locally"""
        if self.kind in (ResourceKind.Absolute, ResourceKind.Relative):
            # todo: I dunno if I wanna deal with relative squirrelyness here
            return True

        elif self.kind is ResourceKind.FullyQualified:
            if self.location_kind in (LocationKind.Local, LocationKind.Archive):
                return True

        return False

    def __fspath__(self):
        if self.is_localized():
            return self._parts.path
        raise UnrootedResource(f"Cannot resolve absolute path: {self.uri}")


class LocalResource(Resource):
    def is_localized(self):
        return True

    def path(self) -> Path:
        return Path(self._parts.path)



from collections.abc import MutableMapping
from .fileio import ManifestMap


class MutableManifestTree(MutableMapping):
    """An abstract (but possibly concrete) representation of a collection of resources
    MMT acts much like a dict, but manages side-effects and invariants

    todo: we may want to lazily eval Resource returns in order to populate the absolute
    path on demand. Or do other side-effecty things. For now, Resource just contains
    redundant info for abspath, for simplicity

    idea: in-place file interfaces to be able to write to virtual "files" in the tree and
    then save the whole manifest. This could be useful for manipulating archives or interacting
    with a remote endpoint.
    """

    def __init__(self, base):
        self._manifest = ManifestMap(base=base, elements={})

    def __getitem__(self, key):
        """Given a relative path, return its Resource"""
        return self._manifest.elements.__getitem__(key)

    def __setitem__(self, key, resource):
        """Insert a resource"""
        self._manifest.elements.__setitem__(key, resource)

    def __delitem__(self, key):
        """Remove a resource"""
        self._manifest.elements.__delitem__(key)

    def __len__(self):
        """Count of files (not dirs) in manifest"""

    def __iter__(self):
        """Iterate over keys"""

    # def add(self, resource):
    #     """Add a resource to the tree"""
    #
    # def discard(self, element):
    #     """Remove a resource"""
    def items(self):
        yield from self._manifest.elements.items()

    def keys(self):
        """Iterates over relative paths"""

        yield from self._manifest.elements.keys()

    def values(self):
        """Iterates over Resource objects"""
        yield from self._manifest.elements.values()
